"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[6749],{5727:function(e){function t(e){e.languages.hpkp={directive:{pattern:/\b(?:includeSubDomains|max-age|pin-sha256|preload|report-to|report-uri|strict)(?=[\s;=]|$)/i,alias:"property"},operator:/=/,punctuation:/;/}}e.exports=t,t.displayName="hpkp",t.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_hpkp.74d3b1ac.chunk.js.map