"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[224],{8347:function(n){function e(n){n.languages["nand2tetris-hdl"]={comment:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,keyword:/\b(?:BUILTIN|CHIP|CLOCKED|IN|OUT|PARTS)\b/,boolean:/\b(?:false|true)\b/,function:/\b[A-Za-z][A-Za-z0-9]*(?=\()/,number:/\b\d+\b/,operator:/=|\.\./,punctuation:/[{}[\];(),:]/}}n.exports=e,e.displayName="nand2tetrisHdl",e.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_nand2tetrisHdl.a6dd0963.chunk.js.map