"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[7176,3047],{8331:function(e,n,t){var a=t(595);function o(e){e.register(a),function(e){e.languages.ejs={delimiter:{pattern:/^<%[-_=]?|[-_]?%>$/,alias:"punctuation"},comment:/^#[\s\S]*/,"language-javascript":{pattern:/[\s\S]+/,inside:e.languages.javascript}},e.hooks.add("before-tokenize",(function(n){e.languages["markup-templating"].buildPlaceholders(n,"ejs",/<%(?!%)[\s\S]+?%>/g)})),e.hooks.add("after-tokenize",(function(n){e.languages["markup-templating"].tokenizePlaceholders(n,"ejs")})),e.languages.eta=e.languages.ejs}(e)}e.exports=o,o.displayName="ejs",o.aliases=["eta"]},595:function(e){function n(e){!function(e){function n(e,n){return"___"+e.toUpperCase()+n+"___"}Object.defineProperties(e.languages["markup-templating"]={},{buildPlaceholders:{value:function(t,a,o,s){if(t.language===a){var r=t.tokenStack=[];t.code=t.code.replace(o,(function(e){if("function"===typeof s&&!s(e))return e;for(var o,i=r.length;-1!==t.code.indexOf(o=n(a,i));)++i;return r[i]=e,o})),t.grammar=e.languages.markup}}},tokenizePlaceholders:{value:function(t,a){if(t.language===a&&t.tokenStack){t.grammar=e.languages[a];var o=0,s=Object.keys(t.tokenStack);!function r(i){for(var u=0;u<i.length&&!(o>=s.length);u++){var l=i[u];if("string"===typeof l||l.content&&"string"===typeof l.content){var g=s[o],c=t.tokenStack[g],p="string"===typeof l?l:l.content,f=n(a,g),k=p.indexOf(f);if(k>-1){++o;var d=p.substring(0,k),m=new e.Token(a,e.tokenize(c,t.grammar),"language-"+a,c),h=p.substring(k+f.length),v=[];d&&v.push.apply(v,r([d])),v.push(m),h&&v.push.apply(v,r([h])),"string"===typeof l?i.splice.apply(i,[u,1].concat(v)):l.content=v}}else l.content&&r(l.content)}return i}(t.tokens)}}}})}(e)}e.exports=n,n.displayName="markupTemplating",n.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_ejs.c3c95368.chunk.js.map