"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[7393],{2195:function(e){function n(e){e.languages.yang={comment:/\/\*[\s\S]*?\*\/|\/\/.*/,string:{pattern:/"(?:[^\\"]|\\.)*"|'[^']*'/,greedy:!0},keyword:{pattern:/(^|[{};\r\n][ \t]*)[a-z_][\w.-]*/i,lookbehind:!0},namespace:{pattern:/(\s)[a-z_][\w.-]*(?=:)/i,lookbehind:!0},boolean:/\b(?:false|true)\b/,operator:/\+/,punctuation:/[{};:]/}}e.exports=n,n.displayName="yang",n.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_yang.b4a63a3b.chunk.js.map