"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[6174],{4988:function(a){function e(a){!function(a){a.languages.false={comment:{pattern:/\{[^}]*\}/},string:{pattern:/"[^"]*"/,greedy:!0},"character-code":{pattern:/'(?:[^\r]|\r\n?)/,alias:"number"},"assembler-code":{pattern:/\d+`/,alias:"important"},number:/\d+/,operator:/[-!#$%&'*+,./:;=>?@\\^_`|~\xdf\xf8]/,punctuation:/\[|\]/,variable:/[a-z]/,"non-standard":{pattern:/[()<BDO\xae]/,alias:"bold"}}}(a)}a.exports=e,e.displayName="$false",e.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_falselang.cfad9d55.chunk.js.map