"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[1438],{7209:function(e){function n(e){e.languages.arff={comment:/%.*/,string:{pattern:/(["'])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:/@(?:attribute|data|end|relation)\b/i,number:/\b\d+(?:\.\d+)?\b/,punctuation:/[{},]/}}e.exports=n,n.displayName="arff",n.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_arff.c5d78406.chunk.js.map