"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[3140],{7761:function(e){function a(e){e.languages.hsts={directive:{pattern:/\b(?:includeSubDomains|max-age|preload)(?=[\s;=]|$)/i,alias:"property"},operator:/=/,punctuation:/;/}}e.exports=a,a.displayName="hsts",a.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_hsts.f4e5dec7.chunk.js.map