"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[4576],{7145:function(n){function e(n){!function(n){n.languages.ignore={comment:/^#.*/m,entry:{pattern:/\S(?:.*(?:(?:\\ )|\S))?/,alias:"string",inside:{operator:/^!|\*\*?|\?/,regex:{pattern:/(^|[^\\])\[[^\[\]]*\]/,lookbehind:!0},punctuation:/\//}}},n.languages.gitignore=n.languages.ignore,n.languages.hgignore=n.languages.ignore,n.languages.npmignore=n.languages.ignore}(n)}n.exports=e,e.displayName="ignore",e.aliases=["gitignore","hgignore","npmignore"]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_ignore.36fffbaa.chunk.js.map