"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[3152],{78:function(e){function o(e){e.languages["go-mod"]=e.languages["go-module"]={comment:{pattern:/\/\/.*/,greedy:!0},version:{pattern:/(^|[\s()[\],])v\d+\.\d+\.\d+(?:[+-][-+.\w]*)?(?![^\s()[\],])/,lookbehind:!0,alias:"number"},"go-version":{pattern:/((?:^|\s)go\s+)\d+(?:\.\d+){1,2}/,lookbehind:!0,alias:"number"},keyword:{pattern:/^([ \t]*)(?:exclude|go|module|replace|require|retract)\b/m,lookbehind:!0},operator:/=>/,punctuation:/[()[\],]/}}e.exports=o,o.displayName="goModule",o.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_goModule.39cf5eaa.chunk.js.map