"use strict";(self.webpackChunkfrontend=self.webpackChunkfrontend||[]).push([[81],{9164:function(n){function t(n){n.languages.properties={comment:/^[ \t]*[#!].*$/m,"attr-value":{pattern:/(^[ \t]*(?:\\(?:\r\n|[\s\S])|[^\\\s:=])+(?: *[=:] *(?! )| ))(?:\\(?:\r\n|[\s\S])|[^\\\r\n])+/m,lookbehind:!0},"attr-name":/^[ \t]*(?:\\(?:\r\n|[\s\S])|[^\\\s:=])+(?= *[=:]| )/m,punctuation:/[=:]/}}n.exports=t,t.displayName="properties",t.aliases=[]}}]);
//# sourceMappingURL=react-syntax-highlighter_languages_refractor_properties.7d3c7fbe.chunk.js.map